@@ -13,11 +13,11 @@
 
 #include "bcachefs.h"
 
-#include "alloc/background.h"
 #include "alloc/backpointers.h"
 #include "alloc/buckets_waiting_for_journal.h"
 #include "alloc/buckets.h"
 #include "alloc/check.h"
+#include "alloc/discard.h"
 #include "alloc/disk_groups.h"
 #include "alloc/foreground.h"
 
@@ -26,7 +26,8 @@
 #include "btree/check.h"
 
 #include "data/copygc.h"
-#include "data/ec.h"
+#include "data/ec/create.h"
+#include "data/ec/init.h"
 #include "data/nocow_locking.h"
 #include "data/write.h"
 
@@ -211,8 +212,7 @@ static inline bool may_alloc_bucket(struct bch_fs *c,
 
 static struct open_bucket *__try_alloc_bucket(struct bch_fs *c,
 					      struct alloc_request *req,
-					      u64 bucket, u8 gen,
-					      struct closure *cl)
+					      u64 bucket, u8 gen)
 {
 	struct bch_dev *ca = req->ca;
 
@@ -227,11 +227,17 @@ static struct open_bucket *__try_alloc_bucket(struct bch_fs *c,
 	guard(spinlock)(&c->allocator.freelist_lock);
 
 	if (unlikely(c->allocator.open_buckets_nr_free <= bch2_open_buckets_reserved(req->watermark))) {
-		if (cl)
-			closure_wait(&c->allocator.open_buckets_wait, cl);
-
 		track_event_change(&c->times[BCH_TIME_blocked_allocate_open_bucket], true);
-		return ERR_PTR(bch_err_throw(c, open_buckets_empty));
+
+		int ret;
+		if (req->cl) {
+			closure_wait(&c->allocator.open_buckets_wait, req->cl);
+			ret = bch_err_throw(c, open_bucket_alloc_blocked);
+		} else {
+			ret = bch_err_throw(c, open_buckets_empty);
+		}
+
+		return ERR_PTR(ret);
 	}
 
 	/* Recheck under lock: */
@@ -261,8 +267,7 @@ static struct open_bucket *__try_alloc_bucket(struct bch_fs *c,
 
 static struct open_bucket *try_alloc_bucket(struct btree_trans *trans,
 					    struct alloc_request *req,
-					    struct btree_iter *freespace_iter,
-					    struct closure *cl)
+					    struct btree_iter *freespace_iter)
 {
 	struct bch_fs *c = trans->c;
 	u64 b = freespace_iter->pos.offset & ~(~0ULL << 56);
@@ -277,7 +282,7 @@ static struct open_bucket *try_alloc_bucket(struct btree_trans *trans,
 	if (ret)
 		return NULL;
 
-	return __try_alloc_bucket(c, req, b, gen, cl);
+	return __try_alloc_bucket(c, req, b, gen);
 }
 
 /*
@@ -285,8 +290,7 @@ static struct open_bucket *try_alloc_bucket(struct btree_trans *trans,
  */
 static noinline struct open_bucket *
 bch2_bucket_alloc_early(struct btree_trans *trans,
-			struct alloc_request *req,
-			struct closure *cl)
+			struct alloc_request *req)
 {
 	struct bch_fs *c = trans->c;
 	struct bch_dev *ca = req->ca;
@@ -348,7 +352,7 @@ again:
 			req->counters.buckets_seen++;
 
 			ob = may_alloc_bucket(c, req, k.k->p)
-				? __try_alloc_bucket(c, req, k.k->p.offset, a->gen, cl)
+				? __try_alloc_bucket(c, req, k.k->p.offset, a->gen)
 				: NULL;
 			if (ob)
 				break;
@@ -369,8 +373,7 @@ again:
 }
 
 static struct open_bucket *bch2_bucket_alloc_freelist(struct btree_trans *trans,
-						      struct alloc_request *req,
-						      struct closure *cl)
+						      struct alloc_request *req)
 {
 	struct bch_dev *ca = req->ca;
 	struct bkey_s_c k;
@@ -411,7 +414,7 @@ again:
 				goto next;
 			}
 
-			ob = try_alloc_bucket(trans, req, &iter, cl);
+			ob = try_alloc_bucket(trans, req, &iter);
 			if (ob) {
 				if (!IS_ERR(ob))
 					*dev_alloc_cursor = iter.pos.offset;
@@ -444,17 +447,21 @@ fail:
 static noinline void bucket_alloc_to_text(struct printbuf *out,
 					  struct bch_fs *c,
 					  struct alloc_request *req,
-					  struct closure *cl,
 					  struct open_bucket *ob)
 {
-	printbuf_tabstop_push(out, 24);
+	printbuf_tabstop_push(out, 32);
+
+	if (req->ca) {
+		prt_printf(out, "dev\t%s (%u)\n",	req->ca->name, req->ca->dev_idx);
+		prt_printf(out, "avail\t%llu\n",	__dev_buckets_free(req->ca, req->usage, req->watermark));
+	}
 
-	prt_printf(out, "dev\t%s (%u)\n",	req->ca->name, req->ca->dev_idx);
 	prt_printf(out, "watermark\t%s\n",	bch2_watermarks[req->watermark]);
 	prt_printf(out, "data type\t%s\n",	__bch2_data_types[req->data_type]);
-	prt_printf(out, "blocking\t%u\n",	cl != NULL);
+	prt_printf(out, "will_retry_target_devices\t%u\n",	req->will_retry_target_devices);
+	prt_printf(out, "will_retry_all_devices\t%u\n",	req->will_retry_all_devices);
+	prt_printf(out, "blocking\t%u\n", !(req->flags & BCH_WRITE_alloc_nowait));
 	prt_printf(out, "free\t%llu\n",	req->usage.buckets[BCH_DATA_free]);
-	prt_printf(out, "avail\t%llu\n",	__dev_buckets_free(req->ca, req->usage, req->watermark));
 	prt_printf(out, "copygc_wait\t%llu/%lli\n",
 		   bch2_copygc_wait_amount(c),
 		   c->copygc.wait - atomic64_read(&c->io_clock[WRITE].now));
@@ -465,7 +472,7 @@ static noinline void bucket_alloc_to_text(struct printbuf *out,
 	prt_printf(out, "nouse\t%llu\n",	req->counters.skipped_nouse);
 	prt_printf(out, "mi_btree_bitmap\t%llu\n", req->counters.skipped_mi_btree_bitmap);
 
-	if (!IS_ERR(ob))
+	if (!IS_ERR_OR_NULL(ob))
 		prt_printf(out, "allocated\t%llu\n", ob->bucket);
 	else
 		prt_printf(out, "err\t%s\n", bch2_err_str(PTR_ERR(ob)));
@@ -475,28 +482,23 @@ static noinline void bucket_alloc_to_text(struct printbuf *out,
  * bch2_bucket_alloc_trans - allocate a single bucket from a specific device
  * @trans:	transaction object
  * @req:	state for the entire allocation
- * @cl:		if not NULL, closure to be used to wait if buckets not available
- * @nowait:	if true, do not wait for buckets to become available
  *
  * Returns:	an open_bucket on success, or an ERR_PTR() on failure.
  */
 static struct open_bucket *bch2_bucket_alloc_trans(struct btree_trans *trans,
-						   struct alloc_request *req,
-						   struct closure *cl,
-						   bool nowait)
+						   struct alloc_request *req)
 {
 	struct bch_fs *c = trans->c;
 	struct bch_dev *ca = req->ca;
 	struct open_bucket *ob = NULL;
 	bool freespace = READ_ONCE(ca->mi.freespace_initialized);
-	u64 avail;
-	bool waiting = nowait;
+	bool waiting = false;
 
 	req->btree_bitmap = req->data_type == BCH_DATA_btree;
 	memset(&req->counters, 0, sizeof(req->counters));
 again:
 	bch2_dev_usage_read_fast(ca, &req->usage);
-	avail = __dev_buckets_free(ca, req->usage, req->watermark);
+	u64 avail = __dev_buckets_free(ca, req->usage, req->watermark);
 
 	if (req->usage.buckets[BCH_DATA_need_discard] >
 	    min(avail, ca->mi.nbuckets >> 7))
@@ -513,15 +515,24 @@ again:
 		    c->recovery.pass_done < BCH_RECOVERY_PASS_check_allocations)
 			goto alloc;
 
-		if (cl && !waiting) {
-			closure_wait(&c->allocator.freelist_wait, cl);
-			waiting = true;
-			goto again;
-		}
-
 		track_event_change(&c->times[BCH_TIME_blocked_allocate], true);
 
-		ob = ERR_PTR(bch_err_throw(c, freelist_empty));
+		if (req->cl &&
+		    !(req->flags & BCH_WRITE_alloc_nowait) &&
+		    !req->will_retry_target_devices &&
+		    !req->will_retry_all_devices) {
+			if (!waiting) {
+				closure_wait(&c->allocator.freelist_wait, req->cl);
+				waiting = true;
+				goto again;
+			}
+
+			bch2_copygc_wakeup(c);
+			ob = ERR_PTR(bch_err_throw(c, bucket_alloc_blocked));
+		} else {
+			ob = ERR_PTR(bch_err_throw(c, freelist_empty));
+		}
+
 		goto err;
 	}
 
@@ -529,8 +540,8 @@ again:
 		closure_wake_up(&c->allocator.freelist_wait);
 alloc:
 	ob = likely(freespace)
-		? bch2_bucket_alloc_freelist(trans, req, cl)
-		: bch2_bucket_alloc_early(trans, req, cl);
+		? bch2_bucket_alloc_freelist(trans, req)
+		: bch2_bucket_alloc_early(trans, req);
 
 	if (req->counters.need_journal_commit * 2 > avail)
 		bch2_journal_flush_async(&c->journal, NULL);
@@ -552,14 +563,18 @@ err:
 
 	if (!ret) {
 		ob->data_type = req->data_type;
+
 		event_inc_trace(c, bucket_alloc, buf,
-			bucket_alloc_to_text(&buf, c, req, cl, ob));
-	} else if (ret == -BCH_ERR_open_buckets_empty) {
+			bucket_alloc_to_text(&buf, c, req, ob));
+	} else if (ret == -BCH_ERR_open_buckets_empty ||
+		   ret == -BCH_ERR_open_bucket_alloc_blocked) {
 		event_inc_trace(c, open_bucket_alloc_fail, buf,
 			bch2_fs_open_buckets_to_text(&buf, c));
-	} else if (!bch2_err_matches(ret, BCH_ERR_transaction_restart))
+	} else if (!bch2_err_matches(ret, BCH_ERR_transaction_restart) &&
+		   !req->will_retry_target_devices &&
+		   !req->will_retry_all_devices)
 		event_inc_trace(c, bucket_alloc_fail, buf,
-			bucket_alloc_to_text(&buf, c, req, cl, ob));
+			bucket_alloc_to_text(&buf, c, req, ob));
 
 	return ob;
 }
@@ -571,14 +586,14 @@ struct open_bucket *bch2_bucket_alloc(struct bch_fs *c, struct bch_dev *ca,
 {
 	struct open_bucket *ob;
 	struct alloc_request req = {
+		.cl		= cl,
 		.watermark	= watermark,
 		.data_type	= data_type,
 		.ca		= ca,
 	};
 
 	CLASS(btree_trans, trans)(c);
-	lockrestart_do(trans,
-		PTR_ERR_OR_ZERO(ob = bch2_bucket_alloc_trans(trans, &req, cl, false)));
+	lockrestart_do(trans, PTR_ERR_OR_ZERO(ob = bch2_bucket_alloc_trans(trans, &req)));
 	return ob;
 }
 
@@ -692,16 +707,18 @@ static int add_new_bucket(struct bch_fs *c,
 	return 0;
 }
 
-inline int bch2_bucket_alloc_set_trans(struct btree_trans *trans,
-				       struct alloc_request *req,
-				       struct dev_stripe_state *stripe,
-				       struct closure *cl)
+int bch2_bucket_alloc_set_trans(struct btree_trans *trans,
+				struct alloc_request *req,
+				struct dev_stripe_state *stripe)
 {
 	struct bch_fs *c = trans->c;
 	int ret = 0;
 
 	BUG_ON(req->nr_effective >= req->nr_replicas);
 
+	if (req->devs_sorted.nr <= 1)
+		req->will_retry_target_devices = false;
+
 	bch2_dev_alloc_list(c, stripe, &req->devs_may_alloc, &req->devs_sorted);
 
 	darray_for_each(req->devs_sorted, i) {
@@ -711,32 +728,28 @@ inline int bch2_bucket_alloc_set_trans(struct btree_trans *trans,
 
 		if (!req->ca->mi.durability && req->have_cache) {
 			bch2_dev_put(req->ca);
+			req->ca = NULL;
 			continue;
 		}
 
-		struct open_bucket *ob = bch2_bucket_alloc_trans(trans, req, cl,
-							req->flags & BCH_WRITE_alloc_nowait);
+		struct open_bucket *ob = bch2_bucket_alloc_trans(trans, req);
 		if (!IS_ERR(ob))
 			bch2_dev_stripe_increment_inlined(req->ca, stripe, &req->usage);
+
 		bch2_dev_put(req->ca);
+		req->ca = NULL;
 
-		if (IS_ERR(ob)) {
+		if (IS_ERR(ob)) { /* don't squash error */
 			ret = PTR_ERR(ob);
-			if (bch2_err_matches(ret, BCH_ERR_transaction_restart) || cl)
-				break;
-			continue;
-		}
-
-		ret = add_new_bucket(c, req, ob);
-		if (ret)
-			break;
+			if (bch2_err_matches(ret, BCH_ERR_transaction_restart) ||
+			    bch2_err_matches(ret, BCH_ERR_operation_blocked) ||
+			    bch2_err_matches(ret, BCH_ERR_open_buckets_empty))
+				return ret;
+		} else if (add_new_bucket(c, req, ob))
+			return 0;
 	}
 
-	if (ret == 1)
-		return 0;
-	if (ret)
-		return ret;
-	return bch_err_throw(c, insufficient_devices);
+	return ret ?: bch_err_throw(c, insufficient_devices);
 }
 
 /* Allocate from stripes: */
@@ -748,19 +761,12 @@ inline int bch2_bucket_alloc_set_trans(struct btree_trans *trans,
  */
 
 static int bucket_alloc_from_stripe(struct btree_trans *trans,
-				    struct alloc_request *req,
-				    struct closure *cl)
+				    struct alloc_request *req)
 {
 	struct bch_fs *c = trans->c;
 	int ret = 0;
 
-	if (req->ec_replicas < 2)
-		return 0;
-
-	if (ec_open_bucket(c, &req->ptrs))
-		return 0;
-
-	struct ec_stripe_head *h = errptr_try(bch2_ec_stripe_head_get(trans, req, 0, cl));
+	struct ec_stripe_head *h = errptr_try(bch2_ec_stripe_head_get(trans, req, 0));
 	if (!h)
 		return 0;
 
@@ -778,6 +784,11 @@ static int bucket_alloc_from_stripe(struct btree_trans *trans,
 				ec_stripe_new_get(h->s, STRIPE_REF_io);
 
 				ret = add_new_bucket(c, req, ob);
+
+				event_inc_trace(c, bucket_alloc_from_stripe, buf, ({
+					bch2_open_bucket_to_text(&buf, c, ob);
+				}));
+
 				goto out;
 			}
 		}
@@ -870,80 +881,6 @@ static int bucket_alloc_set_partial(struct bch_fs *c,
 	return 0;
 }
 
-static int __open_bucket_add_buckets(struct btree_trans *trans,
-				     struct alloc_request *req,
-				     struct closure *_cl)
-{
-	struct bch_fs *c = trans->c;
-	struct open_bucket *ob;
-	struct closure *cl = NULL;
-	unsigned i;
-	int ret;
-
-	req->devs_may_alloc = target_rw_devs(c, req->wp->data_type, req->target);
-
-	/* Don't allocate from devices we already have pointers to: */
-	darray_for_each(*req->devs_have, i)
-		__clear_bit(*i, req->devs_may_alloc.d);
-
-	open_bucket_for_each(c, &req->ptrs, ob, i)
-		__clear_bit(ob->dev, req->devs_may_alloc.d);
-
-	try(bucket_alloc_set_writepoint(c, req));
-
-	try(bucket_alloc_set_partial(c, req));
-
-	if (req->ec) {
-		ret = bucket_alloc_from_stripe(trans, req, _cl);
-	} else {
-retry_blocking:
-		/*
-		 * Try nonblocking first, so that if one device is full we'll try from
-		 * other devices:
-		 */
-		ret = bch2_bucket_alloc_set_trans(trans, req, &req->wp->stripe, cl);
-		if (ret &&
-		    !bch2_err_matches(ret, BCH_ERR_transaction_restart) &&
-		    !bch2_err_matches(ret, BCH_ERR_insufficient_devices) &&
-		    !cl && _cl) {
-			cl = _cl;
-			goto retry_blocking;
-		}
-	}
-
-	return ret;
-}
-
-static int open_bucket_add_buckets(struct btree_trans *trans,
-				   struct alloc_request *req,
-				   struct closure *cl)
-{
-	int ret;
-
-	if (req->ec && !ec_open_bucket(trans->c, &req->ptrs)) {
-		ret = __open_bucket_add_buckets(trans, req, cl);
-		if (bch2_err_matches(ret, BCH_ERR_transaction_restart) ||
-		    bch2_err_matches(ret, BCH_ERR_operation_blocked) ||
-		    bch2_err_matches(ret, BCH_ERR_freelist_empty) ||
-		    bch2_err_matches(ret, BCH_ERR_open_buckets_empty))
-			return ret;
-
-		if ((req->flags & BCH_WRITE_must_ec) &&
-		    bch2_err_matches(ret, BCH_ERR_insufficient_devices))
-			return ret;
-
-		if (req->nr_effective >= req->nr_replicas)
-			return 0;
-	}
-
-	bool ec = false;
-	swap(ec, req->ec);
-	ret = __open_bucket_add_buckets(trans, req, cl);
-	swap(ec, req->ec);
-
-	return ret < 0 ? ret : 0;
-}
-
 /**
  * should_drop_bucket - check if this is open_bucket should go away
  * @ob:		open_bucket to predicate on
@@ -1210,8 +1147,6 @@ deallocate_extra_replicas(struct bch_fs *c,
 int bch2_alloc_sectors_req(struct btree_trans *trans,
 			   struct alloc_request *req,
 			   struct write_point_specifier write_point,
-			   unsigned nr_replicas_required,
-			   struct closure *cl,
 			   struct write_point **wp_ret)
 {
 	struct bch_fs *c = trans->c;
@@ -1220,63 +1155,95 @@ int bch2_alloc_sectors_req(struct btree_trans *trans,
 	unsigned write_points_nr;
 	int i;
 
-	BUG_ON(!req->nr_replicas || !nr_replicas_required);
+	BUG_ON(!req->nr_replicas);
 retry:
-	req->ptrs.nr		= 0;
-	req->nr_effective	= 0;
-	req->have_cache		= false;
-	write_points_nr		= a->write_points_nr;
+	req->ca				= NULL;
+	req->will_retry_all_devices	= req->target && !(req->flags & BCH_WRITE_only_specified_devs);
+	req->will_retry_target_devices	= !(req->flags & BCH_WRITE_alloc_nowait);
+	req->ptrs.nr			= 0;
+	req->nr_effective		= 0;
+	req->have_cache			= false;
+	write_points_nr			= a->write_points_nr;
 
 	*wp_ret = req->wp = writepoint_find(trans, write_point.v);
 
 	req->data_type		= req->wp->data_type;
 
-	int ret = bch2_trans_relock(trans);
-	if (ret)
-		goto err;
-
 	/* metadata may not allocate on cache devices: */
 	if (req->data_type != BCH_DATA_user)
 		req->have_cache = true;
 
-	if (req->target && !(req->flags & BCH_WRITE_only_specified_devs)) {
-		ret = open_bucket_add_buckets(trans, req, NULL);
-		if (!ret ||
-		    bch2_err_matches(ret, BCH_ERR_transaction_restart))
-			goto alloc_done;
-
-		/* Don't retry from all devices if we're out of open buckets: */
-		if (bch2_err_matches(ret, BCH_ERR_open_buckets_empty)) {
-			int ret2 = open_bucket_add_buckets(trans, req, cl);
-			if (!ret2 ||
-			    bch2_err_matches(ret2, BCH_ERR_transaction_restart) ||
-			    bch2_err_matches(ret2, BCH_ERR_open_buckets_empty)) {
-				ret = ret2;
-				goto alloc_done;
-			}
-		}
-
-		/*
-		 * Only try to allocate cache (durability = 0 devices) from the
-		 * specified target:
-		 */
-		req->have_cache	= true;
-		req->target	= 0;
-
-		ret = open_bucket_add_buckets(trans, req, cl);
-	} else {
-		ret = open_bucket_add_buckets(trans, req, cl);
-	}
-alloc_done:
-	BUG_ON(!ret && req->nr_effective < req->nr_replicas);
-
-	if (ret == -BCH_ERR_insufficient_devices &&
-	    req->nr_effective >= nr_replicas_required)
-		ret = 0;
-
+	int ret = bch2_trans_relock(trans);
 	if (ret)
 		goto err;
 
+	while (1) {
+		req->devs_may_alloc = target_rw_devs(c, req->wp->data_type, req->target);
+
+		/* Don't allocate from devices we already have pointers to: */
+		darray_for_each(*req->devs_have, i)
+			__clear_bit(*i, req->devs_may_alloc.d);
+
+		open_bucket_for_each(c, &req->ptrs, ob, i)
+			__clear_bit(ob->dev, req->devs_may_alloc.d);
+
+		ret =   bucket_alloc_set_writepoint(c, req) ?:
+			bucket_alloc_set_partial(c, req) ?:
+			(req->ec
+			 ? bucket_alloc_from_stripe(trans, req)
+			 : bch2_bucket_alloc_set_trans(trans, req, &req->wp->stripe));
+
+		ret = min(ret, 0); /* We return 1 earlier to terminate allocating */
+
+		if (ret &&
+		    ret != -BCH_ERR_freelist_empty &&
+		    ret != -BCH_ERR_insufficient_devices)
+			goto err;
+
+		if (ret && req->will_retry_all_devices) {
+			/*
+			 * Only try to allocate cache (durability = 0 devices) from the
+			 * specified target:
+			 */
+			req->have_cache			= true;
+			req->target			= 0;
+			req->will_retry_all_devices	= false;
+			continue;
+		}
+
+		if (ret && req->will_retry_target_devices) {
+			/*
+			 * When allocating from a target with multiple devices,
+			 * bch2_bucket_alloc_trans() won't block until we've
+			 * attempted all devices in the target once
+			 */
+			req->will_retry_target_devices = false;
+			continue;
+		}
+
+		if (req->nr_effective < req->nr_replicas && req->ec) {
+			if ((req->flags & BCH_WRITE_must_ec)) {
+				ret = bch_err_throw(c, ec_alloc_failed);
+				goto err;
+			}
+
+			req->ec				= false;
+			req->will_retry_target_devices	= true;
+			req->will_retry_all_devices	= req->target && !(req->flags & BCH_WRITE_only_specified_devs);
+			continue;
+		}
+
+		if (ret == -BCH_ERR_insufficient_devices &&
+		    req->nr_effective)
+			ret = 0;
+
+		if (ret)
+			goto err;
+
+		BUG_ON(!req->nr_effective);
+		break;
+	}
+
 	if (req->ec &&
 	    (req->flags & BCH_WRITE_must_ec) &&
 	    !ec_open_bucket(c, &req->ptrs)) {
@@ -1312,6 +1279,7 @@ alloc_done:
 		req->wp->sectors_free = min(req->wp->sectors_free, ob->sectors_free);
 	}
 
+	req->wp->prev_sectors_free = req->wp->sectors_free;
 	req->wp->sectors_free = rounddown(req->wp->sectors_free, block_sectors(c));
 
 	/* Did alignment use up space in an open_bucket? */
@@ -1333,18 +1301,12 @@ err:
 
 	mutex_unlock(&req->wp->lock);
 
-	if (bch2_err_matches(ret, BCH_ERR_freelist_empty) &&
-	    try_decrease_writepoints(trans, write_points_nr))
+	if ((bch2_err_matches(ret, BCH_ERR_freelist_empty) ||
+	     bch2_err_matches(ret, BCH_ERR_bucket_alloc_blocked)) &&
+	    try_decrease_writepoints(trans, write_points_nr)) {
+		if (bch2_err_matches(ret, BCH_ERR_bucket_alloc_blocked))
+			closure_wake_up(&c->allocator.freelist_wait);
 		goto retry;
-
-	if (cl && bch2_err_matches(ret, BCH_ERR_open_buckets_empty))
-		ret = bch_err_throw(c, bucket_alloc_blocked);
-
-	if (cl && !(req->flags & BCH_WRITE_alloc_nowait) &&
-	    bch2_err_matches(ret, BCH_ERR_freelist_empty)) {
-		bch2_copygc_wakeup(c);
-
-		ret = bch_err_throw(c, bucket_alloc_blocked);
 	}
 
 	return ret;
@@ -1522,14 +1484,20 @@ void bch2_fs_open_buckets_to_text(struct printbuf *out, struct bch_fs *c)
 	unsigned nr[BCH_DATA_NR];
 	memset(nr, 0, sizeof(nr));
 
-	for (unsigned i = 0; i < ARRAY_SIZE(a->open_buckets); i++)
-		nr[a->open_buckets[i].data_type]++;
+	for (struct open_bucket *ob = a->open_buckets;
+	     ob < a->open_buckets + ARRAY_SIZE(a->open_buckets);
+	     ob++)
+		if (atomic_read(&ob->pin))
+			nr[ob->data_type]++;
 
-	prt_printf(out, "open buckets allocated\t%i\n",		OPEN_BUCKETS_COUNT - c->allocator.open_buckets_nr_free);
+	prt_printf(out, "open buckets allocated\t%i\n",		OPEN_BUCKETS_COUNT - a->open_buckets_nr_free);
 	prt_printf(out, "open buckets total\t%u\n",		OPEN_BUCKETS_COUNT);
-	prt_printf(out, "open_buckets_btree\t%u\n",		nr[BCH_DATA_btree]);
-	prt_printf(out, "open_buckets_user\t%u\n",		nr[BCH_DATA_user]);
-	prt_printf(out, "open_buckets_wait\t%s\n",		c->allocator.open_buckets_wait.list.first ? "waiting" : "empty");
+
+	for (unsigned i = 0; i < ARRAY_SIZE(nr); i++)
+		if (nr[i])
+			prt_printf(out, "open_buckets %s:\t%u\n", __bch2_data_types[i], nr[i]);
+
+	prt_printf(out, "open_buckets_wait\t%s\n",		a->open_buckets_wait.list.first ? "waiting" : "empty");
 }
 
 void bch2_fs_alloc_debug_to_text(struct printbuf *out, struct bch_fs *c)
@@ -1552,8 +1520,6 @@ void bch2_fs_alloc_debug_to_text(struct printbuf *out, struct bch_fs *c)
 	prt_printf(out, "freelist_wait\t%s\n",			a->freelist_wait.list.first ? "waiting" : "empty");
 	prt_printf(out, "btree reserve cache\t%u\n",		c->btree.reserve_cache.nr);
 	prt_newline(out);
-
-	bch2_fs_open_buckets_to_text(out, c);
 }
 
 void bch2_dev_alloc_debug_to_text(struct printbuf *out, struct bch_dev *ca)
@@ -1589,6 +1555,7 @@ void bch2_dev_alloc_debug_to_text(struct printbuf *out, struct bch_dev *ca)
 
 static noinline void bch2_print_allocator_stuck(struct bch_fs *c, enum bch_watermark watermark)
 {
+	struct bch_fs_allocator *a = &c->allocator;
 	CLASS(printbuf, buf)();
 
 	prt_printf(&buf, "Allocator stuck? Waited for %u seconds, watermark %s\n",
@@ -1602,16 +1569,21 @@ static noinline void bch2_print_allocator_stuck(struct bch_fs *c, enum bch_water
 
 	bch2_printbuf_make_room(&buf, 4096);
 
-	scoped_guard(rcu) {
-		guard(printbuf_atomic)(&buf);
-		for_each_online_member_rcu(c, ca) {
-			prt_printf(&buf, "Dev %u:\n", ca->dev_idx);
-			scoped_guard(printbuf_indent, &buf)
-				bch2_dev_alloc_debug_to_text(&buf, ca);
-			prt_newline(&buf);
+	if (a->freelist_wait.list.first) {
+		scoped_guard(rcu) {
+			guard(printbuf_atomic)(&buf);
+			for_each_online_member_rcu(c, ca) {
+				prt_printf(&buf, "Dev %u:\n", ca->dev_idx);
+				scoped_guard(printbuf_indent, &buf)
+					bch2_dev_alloc_debug_to_text(&buf, ca);
+				prt_newline(&buf);
+			}
 		}
 	}
 
+	if (a->open_buckets_wait.list.first)
+		bch2_fs_open_buckets_to_text(&buf, c);
+
 	prt_printf(&buf, "Copygc debug:\n");
 	scoped_guard(printbuf_indent, &buf)
 		bch2_copygc_wait_to_text(&buf, c);
