@@ -9,6 +9,7 @@
 #include "bcachefs.h"
 
 #include "alloc/buckets.h"
+#include "alloc/disk_groups.h"
 
 #include "btree/bkey_methods.h"
 #include "btree/cache.h"
@@ -19,7 +20,7 @@
 #include "data/checksum.h"
 #include "data/compress.h"
 #include "data/extents.h"
-#include "data/reconcile.h"
+#include "data/reconcile/trigger.h"
 
 #include "fs/inode.h"
 
@@ -131,16 +132,11 @@ static inline int dev_failed(struct bch_dev *ca)
 static inline bool ptr_better(struct bch_fs *c,
 			      const struct extent_ptr_decoded p1,
 			      u64 p1_latency,
-			      struct bch_dev *ca1,
 			      const struct extent_ptr_decoded p2,
-			      u64 p2_latency)
+			      u64 p2_latency,
+			      unsigned preferred_dev,
+			      enum bch_read_flags flags)
 {
-	struct bch_dev *ca2 = bch2_dev_rcu_noerror(c, p2.ptr.dev);
-
-	int failed_delta = dev_failed(ca1) - dev_failed(ca2);
-	if (unlikely(failed_delta))
-		return failed_delta < 0;
-
 	if (static_branch_unlikely(&bch2_force_reconstruct_read))
 		return p1.do_ec_reconstruct > p2.do_ec_reconstruct;
 
@@ -166,6 +162,19 @@ static inline bool ptr_better(struct bch_fs *c,
 	if (unlikely(delta))
 		return delta > 0;
 
+	if (unlikely(flags & BCH_READ_soft_require_read_device)) {
+		if (p1.ptr.dev == preferred_dev) {
+			if (bch2_dev_rotational(c, p2.ptr.dev))
+				return true;
+			p1_latency /= 256;
+		}
+		if (p2.ptr.dev == preferred_dev) {
+			if (bch2_dev_rotational(c, p1.ptr.dev))
+				return false;
+			p2_latency /= 256;
+		}
+	}
+
 	/* Pick at random, biased in favor of the faster device: */
 
 	return bch2_get_random_u64_below(p1_latency + p2_latency) > p1_latency;
@@ -179,7 +188,8 @@ static inline bool ptr_better(struct bch_fs *c,
 int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 			       struct bch_io_failures *failed,
 			       struct extent_ptr_decoded *pick,
-			       int dev)
+			       unsigned preferred_dev,
+			       enum bch_read_flags flags)
 {
 	bool have_csum_errors = false, have_io_errors = false, have_missing_devs = false;
 	bool have_dirty_ptrs = false, have_pick = false;
@@ -206,7 +216,8 @@ int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 		}
 
 		/* Are we being asked to read from a specific device? */
-		if (dev >= 0 && p.ptr.dev != dev)
+		if ((flags & BCH_READ_hard_require_read_device) &&
+		    p.ptr.dev != preferred_dev)
 			continue;
 
 		struct bch_dev *ca = bch2_dev_rcu_noerror(c, p.ptr.dev);
@@ -260,8 +271,9 @@ int bch2_bkey_pick_read_device(struct bch_fs *c, struct bkey_s_c k,
 
 		if (!have_pick ||
 		    ptr_better(c,
-			       p, p_latency, ca,
-			       *pick, pick_latency)) {
+			       p, p_latency,
+			       *pick, pick_latency,
+			       preferred_dev, flags)) {
 			*pick = p;
 			pick_latency = p_latency;
 			have_pick = true;
@@ -809,14 +821,35 @@ unsigned bch2_bkey_replicas(struct bch_fs *c, struct bkey_s_c k)
 	return replicas;
 }
 
-unsigned bch2_extent_ptr_desired_durability(struct bch_fs *c, struct extent_ptr_decoded *p)
+unsigned bch2_dev_durability(struct bch_fs *c, unsigned dev)
 {
-	struct bch_dev *ca = bch2_dev_rcu_noerror(c, p->ptr.dev);
+	guard(rcu)();
+	struct bch_dev *ca = bch2_dev_rcu_noerror(c, dev);
+
+	return ca && ca->mi.state != BCH_MEMBER_STATE_evacuating
+		? ca->mi.durability
+		: 0;
+}
+
+static inline unsigned __extent_ptr_durability(struct bch_dev *ca, struct extent_ptr_decoded *p)
+{
+	if (p->ptr.cached)
+		return 0;
+
+	return p->has_ec
+		? p->ec.redundancy + 1
+		: ca->mi.durability;
+}
+
+int bch2_extent_ptr_desired_durability(struct btree_trans *trans, struct extent_ptr_decoded *p)
+{
+	guard(rcu)();
+	struct bch_dev *ca = bch2_dev_rcu_noerror(trans->c, p->ptr.dev);
 
 	return ca ? __extent_ptr_durability(ca, p) : 0;
 }
 
-unsigned bch2_extent_ptr_durability(struct bch_fs *c, struct extent_ptr_decoded *p)
+static unsigned bch2_extent_ptr_durability_rcu(struct bch_fs *c, struct extent_ptr_decoded *p)
 {
 	struct bch_dev *ca = bch2_dev_rcu_noerror(c, p->ptr.dev);
 
@@ -826,17 +859,72 @@ unsigned bch2_extent_ptr_durability(struct bch_fs *c, struct extent_ptr_decoded
 	return __extent_ptr_durability(ca, p);
 }
 
-unsigned bch2_bkey_durability(struct bch_fs *c, struct bkey_s_c k)
+int bch2_extent_ptr_durability(struct btree_trans *trans, struct extent_ptr_decoded *p)
+{
+	guard(rcu)();
+	return bch2_extent_ptr_durability_rcu(trans->c, p);
+}
+
+int bch2_bkey_durability(struct btree_trans *trans, struct bkey_s_c k, struct bkey_durability *ret)
+{
+	struct bch_fs *c = trans->c;
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	*ret = (struct bkey_durability) {};
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev == BCH_SB_MEMBER_INVALID)
+			continue;
+
+		int d = bch2_extent_ptr_durability(trans, &p);
+		if (d < 0)
+			return d;
+
+		if (test_bit(p.ptr.dev, c->devs_online.d))
+			ret->online += d;
+		ret->total += d;
+	}
+	return 0;
+}
+
+struct bkey_durability bch2_btree_ptr_durability(struct bch_fs *c, struct bkey_s_c k)
+{
+	BUG_ON(!bkey_is_btree_ptr(k.k));
+
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	struct bkey_durability ret = {};
+
+	guard(rcu)();
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev == BCH_SB_MEMBER_INVALID)
+			continue;
+
+		unsigned d = bch2_extent_ptr_durability_rcu(c, &p);
+
+		if (test_bit(p.ptr.dev, c->devs_online.d))
+			ret.online += d;
+		ret.total += d;
+	}
+	return ret;
+}
+
+bool bch2_bkey_can_read(const struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
 	const union bch_extent_entry *entry;
 	struct extent_ptr_decoded p;
-	unsigned durability = 0;
 
-	guard(rcu)();
 	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
-		durability += bch2_extent_ptr_durability(c, &p);
-	return durability;
+		if (!p.ptr.cached &&
+		    (p.ptr.dev != BCH_SB_MEMBER_INVALID ||
+		     p.has_ec))
+			return true;
+
+	return false;
 }
 
 static unsigned bch2_bkey_durability_safe(struct bch_fs *c, struct bkey_s_c k)
@@ -849,7 +937,7 @@ static unsigned bch2_bkey_durability_safe(struct bch_fs *c, struct bkey_s_c k)
 	guard(rcu)();
 	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
 		if (p.ptr.dev < c->sb.nr_devices && c->devs[p.ptr.dev])
-			durability += bch2_extent_ptr_durability(c, &p);
+			durability += bch2_extent_ptr_durability_rcu(c, &p);
 	return durability;
 }
 
@@ -978,11 +1066,6 @@ void bch2_bkey_drop_ptr(const struct bch_fs *c, struct bkey_s k, struct bch_exte
 	}
 
 	bch2_bkey_drop_ptr_noerror(c, k, ptr);
-
-	if (!bch2_bkey_nr_dirty_ptrs(c, k.s_c)) {
-		k.k->type = KEY_TYPE_error;
-		set_bkey_val_u64s(k.k, 0);
-	}
 }
 
 void bch2_bkey_drop_ptrs_mask(const struct bch_fs *c, struct bkey_i *k, unsigned ptrs)
@@ -1048,6 +1131,22 @@ const struct bch_extent_ptr *bch2_bkey_has_device_c(const struct bch_fs *c, stru
 	return NULL;
 }
 
+bool bch2_bkey_has_device_decode(const struct bch_fs *c, struct bkey_s_c k, unsigned dev,
+				 struct extent_ptr_decoded *ret)
+{
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry)
+		if (p.ptr.dev == dev) {
+			*ret = p;
+			return true;
+		}
+
+	return false;
+}
+
 bool bch2_bkey_devs_rw(struct bch_fs *c, struct bkey_s_c k)
 {
 	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
@@ -1069,7 +1168,7 @@ bool bch2_bkey_has_target(struct bch_fs *c, struct bkey_s_c k, unsigned target)
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr)
-		if (bch2_dev_in_target(c, ptr->dev, target) &&
+		if (bch2_dev_in_target_rcu(c, ptr->dev, target) &&
 		    (ca = bch2_dev_rcu_noerror(c, ptr->dev)) &&
 		    (!ptr->cached ||
 		     !dev_ptr_stale_rcu(ca, ptr)))
@@ -1084,12 +1183,22 @@ bool bch2_bkey_in_target(struct bch_fs *c, struct bkey_s_c k, unsigned target)
 
 	guard(rcu)();
 	bkey_for_each_ptr(ptrs, ptr)
-		if (!bch2_dev_in_target(c, ptr->dev, target))
+		if (!bch2_dev_in_target_rcu(c, ptr->dev, target))
 			return false;
 
 	return true;
 }
 
+bool bch2_bkey_has_dev_bad_or_evacuating(struct bch_fs *c, struct bkey_s_c k)
+{
+	guard(rcu)();
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	bkey_for_each_ptr(ptrs, ptr)
+		if (bch2_dev_bad_or_evacuating_rcu(c, ptr->dev))
+			return true;
+	return false;
+}
+
 bool bch2_bkey_matches_ptr(struct bch_fs *c, struct bkey_s_c k,
 			   struct bch_extent_ptr m, u64 offset)
 {
@@ -1178,17 +1287,41 @@ bch2_extent_has_ptr(const struct bch_fs *c, struct bkey_s_c k1,
 	return NULL;
 }
 
-static bool want_cached_ptr(struct bch_fs *c, struct bch_inode_opts *opts,
-			    struct bch_extent_ptr *ptr)
+static bool drop_cached_pointer_trace(struct bch_fs *c, struct bkey_s k,
+				      struct bch_extent_ptr *ptr, const char *msg)
 {
-	unsigned target = opts->promote_target ?: opts->foreground_target;
+	event_add_trace(c, cached_ptr_drop, k.k->size, buf, ({
+		prt_str(&buf, msg);
+		prt_str(&buf, "\ndropping pointer\n  ");
+		bch2_extent_ptr_to_text(&buf, c, ptr);
+		prt_str(&buf, "\nfrom:\n");
+		bch2_bkey_val_to_text(&buf, c, k.s_c);
+	}));
 
-	if (target && !bch2_dev_in_target(c, ptr->dev, target))
-		return false;
+	bch2_bkey_drop_ptr_noerror(c, k, ptr);
+	return true;
+}
 
-	struct bch_dev *ca = bch2_dev_rcu_noerror(c, ptr->dev);
+static bool maybe_drop_cached_ptr(struct bch_fs *c, struct bch_inode_opts *opts,
+				  struct bkey_s k,
+				  struct bch_extent_ptr *ptr, bool have_cached_ptr)
+{
+	if (ptr->cached) {
+		if (have_cached_ptr)
+			return drop_cached_pointer_trace(c, k, ptr, "extent already has another cached pointer");
 
-	return ca && bch2_dev_is_healthy(ca) && !dev_ptr_stale_rcu(ca, ptr);
+		struct bch_dev *ca = bch2_dev_rcu_noerror(c, ptr->dev);
+		if (ca && dev_ptr_stale_rcu(ca, ptr))
+			return drop_cached_pointer_trace(c, k, ptr, "pointer is stale");
+		if (!ca || ca->mi.state == BCH_MEMBER_STATE_evacuating)
+			return drop_cached_pointer_trace(c, k, ptr, "device bad or evacuating");
+
+		unsigned target = opts->promote_target ?: opts->foreground_target;
+		if (target && !bch2_dev_in_target_rcu(c, ptr->dev, target))
+			return drop_cached_pointer_trace(c, k, ptr, "incorrect target");
+	}
+
+	return false;
 }
 
 void bch2_extent_ptr_set_cached(struct bch_fs *c,
@@ -1212,20 +1345,18 @@ void bch2_extent_ptr_set_cached(struct bch_fs *c,
 			 * data. Possibly something we can fix in the future?
 			 */
 			if (&entry->ptr == ptr && p.has_ec) {
-				bch2_bkey_drop_ptr_noerror(c, k, ptr);
+				drop_cached_pointer_trace(c, k, ptr, "ptr->cached && ec currently incompatible");
 				return;
 			}
 
-			if (p.ptr.cached) {
-				if (have_cached_ptr || !want_cached_ptr(c, opts, &p.ptr)) {
-					bch2_bkey_drop_ptr_noerror(c, k, &entry->ptr);
-					ptr = NULL;
-					dropped = true;
-					break;
-				}
-
-				have_cached_ptr = true;
+			if (maybe_drop_cached_ptr(c, opts, k, &entry->ptr, have_cached_ptr)) {
+				/* @ptr is being invalidated; we'll have to find it again */
+				ptr = NULL;
+				dropped = true;
+				break;
 			}
+
+			have_cached_ptr |= p.ptr.cached;
 		}
 	} while (dropped);
 
@@ -1236,10 +1367,8 @@ void bch2_extent_ptr_set_cached(struct bch_fs *c,
 				ptr = ptr2;
 	}
 
-	if (have_cached_ptr || !want_cached_ptr(c, opts, ptr))
-		bch2_bkey_drop_ptr_noerror(c, k, ptr);
-	else
-		ptr->cached = true;
+	ptr->cached = true;
+	maybe_drop_cached_ptr(c, opts, k, ptr, have_cached_ptr);
 }
 
 static bool bch2_bkey_has_stale_ptrs(struct bch_fs *c, struct bkey_s_c k)
@@ -1300,45 +1429,69 @@ void bch2_bkey_drop_extra_cached_ptrs(struct bch_fs *c,
 		bool have_cached_ptr = false;
 		dropped = false;
 
-		bkey_for_each_ptr(ptrs, ptr)
-			if (ptr->cached) {
-				if (have_cached_ptr || !want_cached_ptr(c, opts, ptr)) {
-					bch2_bkey_drop_ptr_noerror(c, k, ptr);
-					dropped = true;
-					break;
-				}
-				have_cached_ptr = true;
-			}
-	} while (dropped);
-}
-
-void bch2_bkey_drop_extra_durability(struct bch_fs *c,
-				     struct bch_inode_opts *opts,
-				     struct bkey_s k)
-{
-	guard(rcu)();
-	unsigned durability = bch2_bkey_durability(c, k.s_c);
-	bool dropped;
-
-	do {
-		union bch_extent_entry *entry;
-		struct extent_ptr_decoded p;
-		dropped = false;
-
-		bkey_for_each_ptr_decode(k.k, bch2_bkey_ptrs(k), p, entry) {
-			unsigned ptr_durability = bch2_extent_ptr_durability(c, &p);
-
-			if (!p.ptr.cached &&
-			    durability - ptr_durability >= opts->data_replicas) {
-				bch2_extent_ptr_set_cached(c, opts, k, &entry->ptr);
-				durability -= ptr_durability;
+		bkey_for_each_ptr(ptrs, ptr) {
+			if (maybe_drop_cached_ptr(c, opts, k, ptr, have_cached_ptr)) {
 				dropped = true;
 				break;
 			}
+
+			have_cached_ptr |= ptr->cached;
 		}
 	} while (dropped);
 }
 
+int bch2_bkey_drop_extra_durability(struct btree_trans *trans,
+				    struct bch_inode_opts *opts,
+				    struct bkey_s k)
+{
+	struct bch_fs *c = trans->c;
+	u8 ptrs_kill = 0;
+	u8 ptr_durability[BCH_BKEY_PTRS_MAX];
+	unsigned durability = 0, nr_ptrs = 0;
+
+	memset(ptr_durability, 0, sizeof(ptr_durability));
+
+	union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+
+	bkey_for_each_ptr_decode(k.k, bch2_bkey_ptrs(k), p, entry) {
+		BUG_ON(nr_ptrs >= ARRAY_SIZE(ptr_durability));
+
+		int d = bch2_extent_ptr_durability(trans, &p);
+		if (d < 0)
+			return d;
+
+		BUG_ON(d > U8_MAX);
+
+		if (!d && !p.ptr.cached)
+			ptrs_kill |= BIT(nr_ptrs);
+
+		ptr_durability[nr_ptrs++] = d;
+		durability += d;
+	}
+
+	for (unsigned i = 0; i < nr_ptrs; i++)
+		if (ptr_durability[i] &&
+		    durability - ptr_durability[i] >= opts->data_replicas) {
+			durability -= ptr_durability[i];
+			ptr_durability[i] = 0;
+			ptrs_kill |= BIT(i);
+		}
+
+	while (ptrs_kill) {
+		unsigned drop = __fls(ptrs_kill);
+		ptrs_kill &= ~BIT(drop);
+
+		bkey_for_each_ptr(bch2_bkey_ptrs(k), ptr)
+			if (!drop--) {
+				bch2_extent_ptr_set_cached(c, opts, k, ptr);
+				break;
+			}
+	}
+
+	return 0;
+}
+
 void bch2_extent_ptr_to_text(struct printbuf *out, struct bch_fs *c, const struct bch_extent_ptr *ptr)
 {
 	struct bch_dev *ca = c ? bch2_dev_rcu_noerror(c, ptr->dev) : NULL;
@@ -1519,7 +1672,7 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	unsigned nonce = UINT_MAX;
 	unsigned nr_ptrs = 0;
 	bool have_written = false, have_unwritten = false, have_ec = false, crc_since_last_ptr = false;
-	bool have_inval_dev_ptrs = false, have_non_inval_dev_ptrs = false;
+	bool have_non_inval_dev_ptrs = false;
 	int ret = 0;
 
 	if (bkey_is_btree_ptr(k.k))
@@ -1528,10 +1681,10 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	bkey_extent_entry_for_each(ptrs, entry) {
 		unsigned type = extent_entry_type(entry);
 
-		bkey_fsck_err_on(type >= c->extent_types_known,
+		bkey_fsck_err_on(type >= c->sb.extent_types_known,
 				 c, extent_ptrs_invalid_entry,
 				 "invalid extent entry type (got %u, max %u)",
-				 type, c->extent_types_known);
+				 type, c->sb.extent_types_known);
 
 		bkey_fsck_err_on(bkey_is_btree_ptr(k.k) &&
 				 type < BCH_EXTENT_ENTRY_MAX &&
@@ -1555,9 +1708,7 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 			have_ec = false;
 			crc_since_last_ptr = false;
 
-			if (entry->ptr.dev == BCH_SB_MEMBER_INVALID)
-				have_inval_dev_ptrs = true;
-			else
+			if (entry->ptr.dev != BCH_SB_MEMBER_INVALID)
 				have_non_inval_dev_ptrs = true;
 
 			nr_ptrs++;
@@ -1646,7 +1797,7 @@ int bch2_bkey_ptrs_validate(struct bch_fs *c, struct bkey_s_c k,
 	 * other fields in bch_btree_ptr_v2
 	 */
 	bkey_fsck_err_on(!bkey_is_btree_ptr(k.k) &&
-			 have_inval_dev_ptrs && !have_non_inval_dev_ptrs,
+			 !have_non_inval_dev_ptrs && !have_ec,
 			 c, extent_ptrs_all_invalid,
 			 "extent ptrs all to BCH_SB_MEMBER_INVALID");
 fsck_err:
