@@ -11,12 +11,11 @@
 
 #include "data/compress.h"
 #include "data/copygc.h"
-#include "data/ec.h"
 #include "data/extents.h"
 #include "data/keylist.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
-#include "data/reconcile.h"
+#include "data/reconcile/trigger.h"
 #include "data/update.h"
 #include "data/write.h"
 
@@ -181,8 +180,6 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	struct bch_inode_opts opts;
 	try(bch2_bkey_get_io_opts(trans, NULL, k, &opts));
 
-	bch2_bkey_propagate_incompressible(c, insert, bkey_i_to_s_c(&new->k_i));
-
 	/*
 	 * @old: extent that we read from
 	 * @insert: key that we're going to update, initialized from
@@ -201,7 +198,7 @@ static int data_update_index_update_key(struct btree_trans *trans,
 
 	unsigned rewrites_found = 0, ptr_bit = 1;
 	bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry_c) {
-		if ((ptr_bit & u->opts.ptrs_kill ) &&
+		if ((ptr_bit & u->opts.ptrs_kill) &&
 		    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert)))) {
 			if (ptr_bit & u->opts.ptrs_io_error)
 				bch2_bkey_drop_ptr_noerror(c, bkey_i_to_s(insert), ptr);
@@ -214,11 +211,10 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	}
 
 	if (u->opts.ptrs_kill && !rewrites_found) {
-		int durability = bch2_bkey_durability(trans, k);
-		if (durability < 0)
-			return durability;
+		struct bkey_durability durability;
+		try(bch2_bkey_durability(trans, k, &durability));
 
-		if (durability >= opts.data_replicas) {
+		if (durability.total >= opts.data_replicas) {
 			count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert,
 						   "no rewrites found:");
 			bch2_btree_iter_advance(iter);
@@ -263,6 +259,8 @@ static int data_update_index_update_key(struct btree_trans *trans,
 
 	bch2_bkey_drop_extra_cached_ptrs(c, &opts, bkey_i_to_s(insert));
 
+	bch2_bkey_propagate_incompressible(c, insert, k);
+
 	bool should_check_enospc = false;
 	s64 i_sectors_delta = 0, disk_sectors_delta = 0;
 	try(bch2_sum_sector_overwrites(trans, iter, insert,
@@ -526,7 +524,6 @@ int bch2_update_unwritten_extent(struct btree_trans *trans,
 				&update->op.devs_have,
 				update->op.nr_replicas,
 				update->op.nr_replicas,
-				update->op.nr_replicas,
 				update->op.watermark,
 				0, &cl, &wp);
 		if (bch2_err_matches(ret, BCH_ERR_operation_blocked)) {
@@ -795,7 +792,7 @@ static int bch2_can_do_write_btree(struct bch_fs *c,
 		return 0;
 
 	if (!(data_opts->write_flags & BCH_WRITE_only_specified_devs)) {
-		unsigned d = bch2_btree_ptr_durability(c, k);
+		unsigned d = bch2_btree_ptr_durability(c, k).total;
 		if (d < opts->data_replicas &&
 		    d < durability_available_on_target(c, watermark, BCH_DATA_btree, 0, &empty,
 						       data_opts->write_flags & BCH_WRITE_alloc_nowait,
@@ -1055,7 +1052,6 @@ int bch2_data_update_init(struct btree_trans *trans,
 		 */
 		m->op.nr_replicas = max(0, (int) (io_opts->data_replicas - durability_keeping)) +
 			m->opts.extra_replicas;
-		m->op.nr_replicas_required = 1;
 
 		/*
 		 * It might turn out that we don't need any new replicas, if the
