@@ -26,13 +26,47 @@
 	BLK_STS(INVAL)				\
 	BLK_STS(REMOVED)			\
 
-#define BLK_STS(n)								\
+#define BLK_STS(n)				\
 	x(BCH_ERR_blockdev_io_error,	BLK_STS_##n)
 
+#define ZSTD_ERRS()					\
+	ZSTD_error(GENERIC)				\
+	ZSTD_error(prefix_unknown)			\
+	ZSTD_error(version_unsupported)			\
+	ZSTD_error(frameParameter_unsupported)		\
+	ZSTD_error(frameParameter_windowTooLarge)	\
+	ZSTD_error(corruption_detected)			\
+	ZSTD_error(checksum_wrong)			\
+	ZSTD_error(dictionary_corrupted)		\
+	ZSTD_error(dictionary_wrong)			\
+	ZSTD_error(dictionaryCreation_failed)		\
+	ZSTD_error(parameter_unsupported)		\
+	ZSTD_error(parameter_outOfBound)		\
+	ZSTD_error(tableLog_tooLarge)			\
+	ZSTD_error(maxSymbolValue_tooLarge)		\
+	ZSTD_error(maxSymbolValue_tooSmall)		\
+	ZSTD_error(stage_wrong)				\
+	ZSTD_error(init_missing)			\
+	ZSTD_error(memory_allocation)			\
+	ZSTD_error(workSpace_tooSmall)			\
+	ZSTD_error(dstSize_tooSmall)			\
+	ZSTD_error(srcSize_wrong)			\
+	ZSTD_error(dstBuffer_null)			\
+	ZSTD_error(frameIndex_tooLarge)			\
+	ZSTD_error(seekableIO)				\
+	ZSTD_error(dstBuffer_wrong)			\
+	ZSTD_error(srcBuffer_wrong)
+
+#define ZSTD_error(n)					\
+	x(BCH_ERR_zstd_error,	ZSTD_error_##n)
+
 #define BCH_ERRCODES()								\
 	x(EIO,				blockdev_io_error)			\
 	BLK_ERRS()								\
 	x(BCH_ERR_blockdev_io_error,	BLK_STS_UNKNOWN)			\
+	x(EIO,				zstd_error)				\
+	ZSTD_ERRS()								\
+	x(BCH_ERR_zstd_error,		ZSTD_error_unknown)			\
 	x(ERANGE,			ERANGE_option_too_small)		\
 	x(ERANGE,			ERANGE_option_too_big)			\
 	x(ERANGE,			projid_too_big)				\
@@ -165,8 +199,9 @@
 	x(EEXIST,			EEXIST_str_hash_set)			\
 	x(EEXIST,			EEXIST_discard_in_flight_add)		\
 	x(EEXIST,			EEXIST_subvolume_create)		\
-	x(ENOSPC,			open_buckets_empty)			\
-	x(ENOSPC,			freelist_empty)				\
+	x(EAGAIN,			open_buckets_empty)			\
+	x(EAGAIN,			freelist_empty)				\
+	x(ENOSPC,			ec_alloc_failed)			\
 	x(BCH_ERR_freelist_empty,	no_buckets_found)			\
 	x(0,				transaction_restart)			\
 	x(BCH_ERR_transaction_restart,	transaction_restart_fault_inject)	\
@@ -231,8 +266,10 @@
 	x(BCH_ERR_data_update_done,	data_update_done_no_writes_needed)	\
 	x(0,				data_update_fail)			\
 	x(BCH_ERR_data_update_fail,	data_update_fail_would_block)		\
+	x(BCH_ERR_data_update_fail,	data_update_fail_in_flight)		\
 	x(BCH_ERR_data_update_fail,	data_update_fail_no_snapshot)		\
 	x(BCH_ERR_data_update_fail,	data_update_fail_no_rw_devs)		\
+	x(BCH_ERR_data_update_fail,	data_update_fail_need_copygc)		\
 	x(EINVAL,			device_state_not_allowed)		\
 	x(EINVAL,			member_info_missing)			\
 	x(EINVAL,			mismatched_block_size)			\
@@ -259,6 +296,7 @@
 	x(EINVAL,			inode_has_child_snapshot)		\
 	x(EINVAL,			varint_decode_error)			\
 	x(EINVAL,			erasure_coding_found_btree_node)	\
+	x(EINVAL,			erasure_coding_stripe_update_err)	\
 	x(EINVAL,			option_negative)			\
 	x(EINVAL,			topology_repair)			\
 	x(EINVAL,			unaligned_io)				\
@@ -290,6 +328,9 @@
 	x(0,				operation_blocked)			\
 	x(BCH_ERR_operation_blocked,	btree_cache_cannibalize_lock_blocked)	\
 	x(BCH_ERR_operation_blocked,	journal_res_blocked)			\
+	x(BCH_ERR_operation_blocked,	bucket_alloc_blocked)			\
+	x(BCH_ERR_operation_blocked,	open_bucket_alloc_blocked)		\
+	x(BCH_ERR_operation_blocked,	stripe_alloc_blocked)			\
 	x(BCH_ERR_journal_res_blocked,	journal_blocked)			\
 	x(BCH_ERR_journal_res_blocked,	journal_max_in_flight)			\
 	x(BCH_ERR_journal_res_blocked,	journal_max_open)			\
@@ -298,8 +339,6 @@
 	x(BCH_ERR_journal_res_blocked,	journal_buf_enomem)			\
 	x(BCH_ERR_journal_res_blocked,	journal_stuck)				\
 	x(BCH_ERR_journal_res_blocked,	journal_retry_open)			\
-	x(BCH_ERR_journal_res_blocked,	bucket_alloc_blocked)			\
-	x(BCH_ERR_journal_res_blocked,	stripe_alloc_blocked)			\
 	x(BCH_ERR_invalid,		invalid_sb)				\
 	x(BCH_ERR_invalid_sb,		invalid_sb_magic)			\
 	x(BCH_ERR_invalid_sb,		invalid_sb_version)			\
@@ -360,22 +399,23 @@
 	x(EIO,				no_encryption_key)			\
 	x(EIO,				insufficient_journal_devices)		\
 	x(EIO,				device_offline)				\
+	x(EIO,				stripe_create_device_offline)			\
 	x(EIO,				EIO_fault_injected)			\
 	x(EIO,				ec_block_read)				\
 	x(EIO,				ec_block_write)				\
 	x(EIO,				recompute_checksum)			\
-	x(EIO,				decompress)				\
+	x(BCH_ERR_data_read_retry_avoid,decompress)				\
 	x(BCH_ERR_decompress,		decompress_exceeded_max_encoded_extent)	\
 	x(BCH_ERR_decompress,		decompress_lz4)				\
 	x(BCH_ERR_decompress,		decompress_gzip)			\
 	x(BCH_ERR_decompress,		decompress_zstd_src_len_bad)		\
-	x(BCH_ERR_decompress,		decompress_zstd)			\
+	x(BCH_ERR_decompress,		decompress_zstd_size_mismatch)		\
 	x(EIO,				data_write)				\
 	x(BCH_ERR_data_write,		data_write_io)				\
 	x(BCH_ERR_data_write,		data_write_csum)			\
 	x(BCH_ERR_data_write,		data_write_invalid_ptr)			\
 	x(BCH_ERR_data_write,		data_write_misaligned)			\
-	x(BCH_ERR_decompress,		data_read)				\
+	x(EIO,				data_read)				\
 	x(BCH_ERR_data_read,		no_device_to_read_from)			\
 	x(BCH_ERR_data_read,		no_devices_valid)			\
 	x(BCH_ERR_data_read,		data_read_io_err)			\
@@ -402,7 +442,6 @@
 	x(BCH_ERR_nopromote,		nopromote_unwritten)			\
 	x(BCH_ERR_nopromote,		nopromote_congested)			\
 	x(BCH_ERR_nopromote,		nopromote_ratelimited)			\
-	x(BCH_ERR_nopromote,		nopromote_in_flight)			\
 	x(BCH_ERR_nopromote,		nopromote_no_writes)			\
 	x(BCH_ERR_nopromote,		nopromote_enomem)			\
 	x(0,				invalid_snapshot_node)			\
@@ -452,4 +491,8 @@ static inline long bch2_err_class(long err)
 const char *bch2_blk_status_to_str(blk_status_t);
 enum bch_errcode blk_status_to_bch_err(blk_status_t);
 
+#include <linux/zstd_errors.h>
+
+enum bch_errcode zstd_err_to_bch_err(ZSTD_ErrorCode);
+
 #endif /* _BCACHFES_ERRCODE_H */
